from __future__ import annotations

import json
import uuid
from urllib import error, request

from tach.cache.setup import resolve_dot_tach
from tach.filesystem import find_project_config_root

# TODO: don't call find_project_config_root separately in each function, pass as argument


def get_uid() -> uuid.UUID | None:
    project_path = find_project_config_root()
    if project_path is None:
        return
    info_path = project_path / ".tach" / "tach.info"
    if not info_path.exists():
        resolve_dot_tach()
    contents = info_path.read_text().strip()
    uid = uuid.UUID(contents)
    return uid


def get_latest_version() -> str | None:
    project_path = find_project_config_root()
    if project_path is None:
        return
    latest_version_path = project_path / ".tach" / ".latest-version"
    if not latest_version_path.exists():
        return
    version = latest_version_path.read_text().strip()
    return version


def update_latest_version() -> None:
    project_path = find_project_config_root()
    if project_path is None:
        return
    url = "https://pypi.org/pypi/tach/json"
    try:
        # Sending a GET request to the URL
        with request.urlopen(url, timeout=1) as response:
            if response.status == 200:
                data = response.read().decode()
                json_data = json.loads(data)
                latest_version = json_data["info"]["version"]
            else:
                return
    except (error.URLError, KeyError):
        return
    (project_path / ".tach" / ".latest-version").write_text(latest_version)



from myorg import pack_a, pack_d, pack_e
from myorg.pack_f import hello_world


def test_package():
    assert hello_world() == "Hello world from package F!"


def test_integration():
    assert pack_a.hello_world() == "Hello world from package A!"
    assert pack_d.hello_world() == "Hello world from package D!"
    assert pack_e.hello_world() == "Hello world from package E!"



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from unittest.mock import Mock

import pytest

from tach import cli
from tach.extension import ProjectConfig


@dataclass
class ErrorInfo:
    exception_message: str
    dependency_error: bool = False
    interface_error: bool = False
    deprecated: bool = False

    def is_dependency_error(self) -> bool:
        return self.dependency_error

    def is_interface_error(self) -> bool:
        return self.interface_error

    def is_deprecated(self) -> bool:
        return self.deprecated

    def to_pystring(self) -> str:
        return self.exception_message


@dataclass
class BoundaryError:
    file_path: Path
    line_number: int
    import_mod_path: str
    error_info: ErrorInfo


@dataclass
class CheckDiagnostics:
    errors: list[BoundaryError]
    deprecated_warnings: list[BoundaryError]
    warnings: list[str]


@pytest.fixture
def mock_check(mocker) -> Mock:
    mock = Mock(
        return_value=CheckDiagnostics(errors=[], deprecated_warnings=[], warnings=[])
    )  # default to a return with no errors
    mocker.patch("tach.cli.check", mock)
    return mock


@pytest.fixture
def mock_project_config(mocker) -> None:
    def mock_project_config(root: str = "") -> ProjectConfig:
        return ProjectConfig()

    mocker.patch("tach.cli.parse_project_config", mock_project_config)


def test_execute_with_config(capfd, mock_check, mock_project_config):
    # Test with a valid path as mocked
    args, _ = cli.parse_arguments(["check"])
    assert args.command == "check"
    with pytest.raises(SystemExit) as sys_exit:
        cli.tach_check(Path())
    captured = capfd.readouterr()
    assert sys_exit.value.code == 0
    assert "✅" in captured.out
    assert "All modules validated!" in captured.out


def test_execute_with_error(capfd, mock_check, mock_project_config):
    # Mock an error returned from check
    location = Path("valid_dir/file.py")
    message = "Import valid_dir in valid_dir/file.py is blocked by boundary"
    mock_check.return_value = CheckDiagnostics(
        deprecated_warnings=[],
        warnings=[],
        errors=[
            BoundaryError(
                file_path=location,
                line_number=0,
                import_mod_path="valid_dir",
                error_info=ErrorInfo(
                    exception_message="Import valid_dir in valid_dir/file.py is blocked by boundary",
                ),
            )
        ],
    )
    with pytest.raises(SystemExit) as sys_exit:
        cli.tach_check(Path())
    captured = capfd.readouterr()
    assert sys_exit.value.code == 1
    assert str(location) in captured.err
    assert message in captured.err


def test_invalid_command(capfd):
    with pytest.raises(SystemExit) as sys_exit:
        # Test with an invalid command
        cli.parse_arguments(["help"])
    captured = capfd.readouterr()
    assert sys_exit.value.code == 2
    assert "invalid choice: 'help" in captured.err


def test_execute_with_valid_exclude(capfd, mock_check, mock_project_config):
    with pytest.raises(SystemExit) as sys_exit:
        # Test with a valid path as mocked
        args, _ = cli.parse_arguments(["check", "--exclude", "valid_dir"])
        exclude_paths = args.exclude.split(",")
        cli.tach_check(Path(), exclude_paths=exclude_paths)
    captured = capfd.readouterr()
    assert sys_exit.value.code == 0
    assert "✅" in captured.out
    assert "All modules validated!" in captured.out



from __future__ import annotations

from copy import copy
from pathlib import Path
from typing import Any, Protocol

import pytest

from tach import filesystem as fs
from tach.errors import TachSetupError
from tach.extension import TachPytestPluginHandler
from tach.filesystem.git_ops import get_changed_files
from tach.parsing import parse_project_config


class TachConfig(Protocol):
    tach_handler: TachPytestPluginHandler

    def getoption(self, name: str) -> Any: ...


def pytest_addoption(parser: pytest.Parser):
    group = parser.getgroup("tach")
    group.addoption(
        "--tach-base",
        default="main",
        help="Base commit to compare against when determining affected tests [default: main]",
    )
    group.addoption(
        "--tach-head",
        default="",
        help="Head commit to compare against when determining affected tests [default: current filesystem]",
    )


@pytest.hookimpl(tryfirst=True)
def pytest_configure(config: TachConfig):
    project_root = fs.find_project_config_root() or Path.cwd()
    project_config = parse_project_config(root=project_root)
    if project_config is None:
        raise TachSetupError("In Tach pytest plugin: No project config found")

    base = config.getoption("--tach-base")
    head = config.getoption("--tach-head")

    kwargs: dict[str, Any] = {"project_root": project_root}
    if head:
        kwargs["head"] = head
    if base:
        kwargs["base"] = base
    changed_files = get_changed_files(**kwargs)

    # Store the handler instance on the config object so other hooks can access it
    config.tach_handler = TachPytestPluginHandler(
        project_root=project_root,
        project_config=project_config,
        changed_files=changed_files,
        all_affected_modules={changed_file.resolve() for changed_file in changed_files},
    )


def pytest_collection_modifyitems(
    session: pytest.Session,
    config: TachConfig,
    items: list[pytest.Item],
):
    handler = config.tach_handler
    seen: set[Path] = set()
    for item in copy(items):
        if not item.path:
            continue
        if str(item.path) in handler.removed_test_paths:
            handler.num_removed_items += 1
            items.remove(item)
            continue
        if item.path in seen:
            continue

        if str(item.path) in handler.all_affected_modules:
            # If this test file was changed,
            # then we know we need to rerun it
            seen.add(item.path)
            continue

        if handler.should_remove_items(file_path=item.path.resolve()):
            handler.num_removed_items += 1
            items.remove(item)
            handler.remove_test_path(item.path)

        seen.add(item.path)


def pytest_report_collectionfinish(
    config: TachConfig,
    start_path: Path,
    startdir: Any,
    items: list[pytest.Item],
) -> str | list[str]:
    handler = config.tach_handler
    return [
        f"[Tach] Skipped {len(handler.removed_test_paths)} test file{'s' if len(handler.removed_test_paths) > 1 else ''}"
        f" ({handler.num_removed_items} tests)"
        " since they were unaffected by current changes.",
        *(
            f"[Tach] > Skipped '{test_path}'"
            for test_path in handler.removed_test_paths
        ),
    ]


def pytest_terminal_summary(terminalreporter: Any, exitstatus: int, config: TachConfig):
    config.tach_handler.tests_ran_to_completion = True



from __future__ import annotations

import json
import os
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

from tach.errors import TachError, TachSetupError

if TYPE_CHECKING:
    from git import Repo


@dataclass
class GitBranchInfo:
    repo: str
    name: str
    commit: str
    owner: str
    user_name: str
    email: str


def is_github_actions():
    return os.environ.get("GITHUB_ACTIONS") == "true"


def _get_branch_name(repo: Repo) -> str:
    # GHA uses a detached HEAD / shallow clone in actions/checkout@v4
    if is_github_actions():
        event_name = os.environ.get("GITHUB_EVENT_NAME")
        # Different environment variables are used for PRs and pushes
        if event_name == "pull_request":
            return os.environ["GITHUB_HEAD_REF"]
        return os.environ["GITHUB_REF_NAME"]
    else:
        return repo.active_branch.name


def _get_commit(repo: Repo) -> str:
    # GHA uses a detached HEAD / shallow clone in actions/checkout@v4
    if is_github_actions():
        event_name = os.environ.get("GITHUB_EVENT_NAME")
        event_path = os.environ.get("GITHUB_EVENT_PATH")
        if event_path and event_name == "pull_request":
            with open(event_path) as f:
                event_info = json.load(f)
                # Pull commit from the PR event in GHA
                return event_info["pull_request"]["head"]["sha"]
    return repo.head.commit.hexsha


def get_current_branch_info(
    project_root: Path, allow_dirty: bool = False
) -> GitBranchInfo:
    # Local import because git-python takes ~80ms to load
    from git import InvalidGitRepositoryError, NoSuchPathError, Repo

    try:
        repo = Repo(project_root, search_parent_directories=True)
    except (InvalidGitRepositoryError, NoSuchPathError):
        raise TachSetupError("The project does not appear to be a git repository!")

    # Fail if the branch is not clean
    if not allow_dirty and repo.is_dirty():
        raise TachError(
            "The current branch is not clean, please commit your changes before running this command!"
        )

    try:
        # TODO: support slashes or org names
        url_parts = repo.remotes.origin.url.split("/")
        repo_name = url_parts[-1].replace(".git", "")
        owner_name = url_parts[0].split(":")[-1]
        config_reader = repo.config_reader()
        user_name = str(config_reader.get_value("user", "name", default=""))
        email = str(config_reader.get_value("user", "email", default=""))
        branch = _get_branch_name(repo)
        commit = _get_commit(repo)
    except Exception as e:
        raise TachError(f"Failed to determine current branch information!\nError: {e}")

    return GitBranchInfo(
        repo=repo_name,
        owner=owner_name,
        name=branch,
        commit=commit,
        user_name=user_name,
        email=email,
    )


def get_changed_files(
    project_root: Path, head: str = "", base: str = "main"
) -> list[Path]:
    # Local import because git-python takes ~80ms to load
    from git import GitCommandError, InvalidGitRepositoryError, NoSuchPathError, Repo

    try:
        repo = Repo(project_root, search_parent_directories=True)
    except (InvalidGitRepositoryError, NoSuchPathError):
        raise TachSetupError(
            "The project does not appear to be a git repository, cannot determine changed files!"
        )

    try:
        if head:
            diff: str = repo.git.diff("--name-status", head, base)
        else:
            # If head is not provided, we can diff against 'base' from the current filesystem
            diff: str = repo.git.diff("--name-status", base)
    except GitCommandError:
        head_display = f"'{head}'" if head else "current filesystem"
        raise TachError(f"Failed to check diff between '{base}' and {head_display}!")

    diff_lines = diff.splitlines()
    changed_files: set[str] = set()

    for line in diff_lines:
        _, *files = line.split("\t")
        changed_files.update(files)

    if not head:
        # If we are using the current filesystem, there may be relevant changes in untracked files
        untracked_files: str = repo.git.ls_files("--others", "--exclude-standard")
        changed_files.update(untracked_files.splitlines())

    # return list of unique Paths
    git_root: str = repo.git.rev_parse("--show-toplevel")
    return [(Path(git_root) / filepath).resolve() for filepath in changed_files]


__all__ = ["get_changed_files", "get_current_branch_info"]



from __future__ import annotations

import fnmatch
import re
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path


# Assumes 'relative_path' is a path relative to the project root
def is_path_excluded(
    exclude_paths: list[str], relative_path: Path, use_regex_matching: bool
) -> bool:
    if not exclude_paths:
        return False

    # TODO: deprecate regex excludes
    path_for_regex = f"{relative_path}/"
    return any(
        (
            re.match(exclude_path, path_for_regex)
            if use_regex_matching
            else fnmatch.fnmatch(str(relative_path), exclude_path)
        )
        for exclude_path in exclude_paths
    )



from __future__ import annotations

import pytest

from tach.cli import tach_check
from tach.icons import SUCCESS, WARNING


def test_valid_example_dir(example_dir, capfd):
    project_root = example_dir / "valid"
    with pytest.raises(SystemExit) as exc_info:
        tach_check(project_root=project_root)
    assert exc_info.value.code == 0
    captured = capfd.readouterr()
    assert SUCCESS in captured.out  # success state
    assert WARNING in captured.err  # deprecated warning


def test_valid_example_dir_monorepo(example_dir):
    project_root = example_dir / "monorepo"
    with pytest.raises(SystemExit) as exc_info:
        tach_check(project_root=project_root)
    assert exc_info.value.code == 0



from __future__ import annotations

import os
import tempfile
from pathlib import Path

import pytest

from tach.constants import DEFAULT_EXCLUDE_PATHS
from tach.extension import get_project_imports, set_excluded_paths


# Utility function to create temporary files with content
def create_temp_file(directory, filename, content):
    filepath = os.path.join(directory, filename)
    with open(filepath, "w") as f:
        f.write(content)
    return filepath


@pytest.fixture
def temp_project():
    with tempfile.TemporaryDirectory() as project_root:
        project_root = Path(project_root)
        # This is tech debt!!
        set_excluded_paths(
            str(project_root), DEFAULT_EXCLUDE_PATHS, use_regex_matching=True
        )

        # Creating some sample Python files in a nested structure
        (project_root / "a" / "b").mkdir(parents=True, exist_ok=True)
        (project_root / "d" / "e").mkdir(parents=True, exist_ok=True)
        (project_root / "local" / "g" / "h").mkdir(parents=True, exist_ok=True)
        (project_root / "local" / "m" / "n").mkdir(parents=True, exist_ok=True)
        (project_root / "parent").mkdir(parents=True, exist_ok=True)

        # Create __init__.py files in each directory
        (project_root / "a" / "__init__.py").touch()
        (project_root / "a" / "b" / "__init__.py").touch()
        (project_root / "d" / "__init__.py").touch()
        (project_root / "d" / "e" / "__init__.py").touch()
        (project_root / "local" / "__init__.py").touch()
        (project_root / "local" / "g" / "__init__.py").touch()
        (project_root / "local" / "g" / "h" / "__init__.py").touch()
        (project_root / "local" / "m" / "__init__.py").touch()
        (project_root / "local" / "m" / "n" / "__init__.py").touch()
        (project_root / "parent" / "__init__.py").touch()

        file1_content = """
import os
from local.file2 import b
"""
        file2_content = """
from ..file1 import y
"""
        file3_content = """
if TYPE_CHECKING:
    from local.file2 import c
"""
        file4_content = """
# tach-ignore(external dependency)
from a.b import c
from d.e import f  # tach-ignore(legacy import) f
from local.g.h import (  # tach-ignore(deprecated, using j instead) i
    i,
    j,
)

# tach-ignore(temporary workaround) k
from local.m.n import k, l

import file3
"""

        create_temp_file(project_root, "file1.py", file1_content)
        create_temp_file(project_root, "local/file2.py", file2_content)
        create_temp_file(project_root, "file3.py", file3_content)
        create_temp_file(project_root, "file4.py", file4_content)

        yield project_root


def test_regular_imports(temp_project):
    result = get_project_imports(
        [str(temp_project)],
        str(temp_project / "file1.py"),
        ignore_type_checking_imports=True,
        include_string_imports=False,
    )
    expected = [("local.file2.b", 3)]
    assert result == expected


def test_relative_imports(temp_project):
    result = get_project_imports(
        [str(temp_project)],
        str(temp_project / "local/file2.py"),
        ignore_type_checking_imports=True,
        include_string_imports=False,
    )
    expected = [("file1.y", 2)]
    assert result == expected


def test_ignore_type_checking_imports(temp_project):
    result = get_project_imports(
        [str(temp_project)],
        str(temp_project / "file3.py"),
        ignore_type_checking_imports=True,
        include_string_imports=False,
    )
    expected = []
    assert result == expected


def test_include_type_checking_imports(temp_project):
    result = get_project_imports(
        [str(temp_project)],
        str(temp_project / "file3.py"),
        ignore_type_checking_imports=False,
        include_string_imports=False,
    )
    expected = [("local.file2.c", 3)]
    assert result == expected


def test_mixed_imports(temp_project):
    mixed_content = """
import sys
if TYPE_CHECKING:
    from .file2 import c
from ..file1 import x
"""
    create_temp_file(temp_project, "local/file4.py", mixed_content)
    result = get_project_imports(
        [str(temp_project)],
        str(temp_project / "local/file4.py"),
        ignore_type_checking_imports=True,
        include_string_imports=False,
    )
    expected = [("file1.x", 5)]
    assert result == expected

    result = get_project_imports(
        [str(temp_project)],
        str(temp_project / "local/file4.py"),
        ignore_type_checking_imports=False,
        include_string_imports=False,
    )
    expected = [("local.file2.c", 4), ("file1.x", 5)]
    assert result == expected


def test_external_imports(temp_project):
    external_content = """
import os
from external_module import something
"""
    create_temp_file(temp_project, "file5.py", external_content)
    result = get_project_imports(
        [str(temp_project)],
        str(temp_project / "file5.py"),
        ignore_type_checking_imports=True,
        include_string_imports=False,
    )
    expected = []  # 'os' and 'external_module' are not within the project root
    assert result == expected


def test_external_and_internal_imports(temp_project):
    mixed_content = """
import os
from file1 import c
from external_module import something
"""
    create_temp_file(temp_project, "file6.py", mixed_content)
    result = get_project_imports(
        [str(temp_project)],
        str(temp_project / "file6.py"),
        ignore_type_checking_imports=True,
        include_string_imports=False,
    )
    expected = [
        ("file1.c", 3),
    ]
    assert result == expected


def test_ignored_imports(temp_project):
    result = get_project_imports(
        [str(temp_project)],
        str(temp_project / "file4.py"),
        ignore_type_checking_imports=True,
        include_string_imports=False,
    )
    expected = [
        ("local.g.h.j", 7),  # only 'i' is ignored, 'j' is included
        ("local.m.n.l", 11),  # only 'k' is ignored, 'l' is included
        ("file3", 13),
    ]
    assert result == expected


def test_file_outside_source_root(temp_project, tmp_path):
    mixed_content = """
import os
from file1 import c
from external_module import something
"""

    path_outside_source_root = tmp_path / "outside_src_root.py"
    path_outside_source_root.write_text(mixed_content)

    result = get_project_imports(
        [str(temp_project)],
        str(path_outside_source_root),
        ignore_type_checking_imports=True,
        include_string_imports=False,
    )
    expected = [
        ("file1.c", 3),
    ]
    assert result == expected


def test_relative_import_from_parent(temp_project):
    # Create a nested directory structure
    (temp_project / "parent" / "child").mkdir(parents=True, exist_ok=True)

    # Create a file in the parent directory
    parent_file_content = """
def parent_function():
    pass
"""
    create_temp_file(temp_project / "parent", "parent_module.py", parent_file_content)

    # Create a file in the child directory with a relative import from the parent
    child_file_content = """
from .. import parent_module

def child_function():
    parent_module.parent_function()
"""
    create_temp_file(
        temp_project / "parent" / "child", "child_module.py", child_file_content
    )

    result = get_project_imports(
        [str(temp_project)],
        str(temp_project / "parent" / "child" / "child_module.py"),
        ignore_type_checking_imports=True,
        include_string_imports=False,
    )
    expected = [("parent.parent_module", 2)]
    assert result == expected


def test_ignore_comments(temp_project):
    """Test different variations of tach-ignore comments"""
    content = """
# tach-ignore(skip all imports on next line)
from a.b import c, d

from d.e import f  # tach-ignore(deprecated) f
from local.g.h import i, j  # tach-ignore(using j instead) i

# tach-ignore(temporary) k
from local.m.n import k, l
"""
    create_temp_file(temp_project, "ignore_test.py", content)

    result = get_project_imports(
        [str(temp_project)],
        str(temp_project / "ignore_test.py"),
        ignore_type_checking_imports=True,
        include_string_imports=False,
    )
    expected = [
        ("local.g.h.j", 6),  # only 'i' is ignored
        ("local.m.n.l", 9),  # only 'k' is ignored
    ]
    assert result == expected



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from tach.errors import TachError
from tach.extension import check_external_dependencies, set_excluded_paths
from tach.utils.external import get_module_mappings, is_stdlib_module

if TYPE_CHECKING:
    from pathlib import Path

    from tach.extension import ProjectConfig


@dataclass
class ExternalCheckDiagnostics:
    undeclared_dependencies: dict[str, list[str]]
    unused_dependencies: dict[str, list[str]]


def extract_module_mappings(rename: list[str]) -> dict[str, list[str]]:
    try:
        return {
            module: [name] for module, name in [module.split(":") for module in rename]
        }
    except ValueError as e:
        raise TachError(
            "Invalid rename format: expected format is a list of 'module:name' pairs, e.g. ['PIL:pillow']"
        ) from e


def check_external(
    project_root: Path,
    project_config: ProjectConfig,
    exclude_paths: list[str],
) -> ExternalCheckDiagnostics:
    serialized_source_roots = [
        str(project_root / source_root) for source_root in project_config.source_roots
    ]
    set_excluded_paths(
        project_root=str(project_root),
        exclude_paths=exclude_paths,
        use_regex_matching=project_config.use_regex_matching,
    )

    metadata_module_mappings = get_module_mappings()
    if project_config.external.rename:
        metadata_module_mappings.update(
            extract_module_mappings(project_config.external.rename)
        )
    diagnostics = check_external_dependencies(
        project_root=str(project_root),
        source_roots=serialized_source_roots,
        module_mappings=metadata_module_mappings,
        ignore_type_checking_imports=project_config.ignore_type_checking_imports,
    )
    undeclared_dependencies_by_file = diagnostics[0]
    unused_dependencies_by_project = diagnostics[1]

    excluded_external_modules = set(project_config.external.exclude)
    filtered_undeclared_dependencies: dict[str, list[str]] = {}
    for filepath, undeclared_dependencies in undeclared_dependencies_by_file.items():
        dependencies = set(
            filter(
                lambda dependency: not is_stdlib_module(dependency)
                and dependency not in excluded_external_modules,
                undeclared_dependencies,
            )
        )
        if dependencies:
            filtered_undeclared_dependencies[filepath] = list(dependencies)
    filtered_unused_dependencies: dict[str, list[str]] = {}
    for filepath, unused_dependencies in unused_dependencies_by_project.items():
        dependencies = set(
            filter(
                lambda dependency: not is_stdlib_module(dependency)
                and dependency not in excluded_external_modules,
                unused_dependencies,
            )
        )
        if dependencies:
            filtered_unused_dependencies[filepath] = list(dependencies)

    return ExternalCheckDiagnostics(
        undeclared_dependencies=filtered_undeclared_dependencies,
        unused_dependencies=filtered_unused_dependencies,
    )


__all__ = ["check_external"]



from __future__ import annotations

import sys


def start():
    try:
        from tach.cli import main

        main()
    except KeyboardInterrupt:
        print("Exiting...")
        sys.exit(1)


if __name__ == "__main__":
    start()

__all__ = ["start"]



from myorg import pack_a, pack_b
from myorg.pack_c import hello_world


def test_package():
    assert hello_world() == "Hello world from package C!"


def test_integration():
    assert pack_a.hello_world() == "Hello world from package A!"
    assert pack_b.hello_world() == "Hello world from package B!"
